#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <dirent.h>
#include "common.h"

#define TESTFILE "samygo_testfile"
#define LOGBUFSIZE 1024

static int pvr_logfd = -1;

int
pvr_log(char *fmt, ...)
{
	va_list ap;
	char buf[LOGBUFSIZE];
	ssize_t written;
	int len;
	
	if (pvr_logfd < 0)
		return (0);

	va_start(ap, fmt);
	len = vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	if (len < 0)
		return (-1);
	if (len > sizeof(buf) - 1)
		len = sizeof(buf) - 1;
	written = write(pvr_logfd, buf, len);
	if (written != len)
		return (-1);

	return 0;
}

int
pvr_log_init(char *fname)
{
	int fd;

	if ((fd = open(fname, O_WRONLY|O_APPEND|O_CREAT|O_SYNC|O_TRUNC, 0644)) >= 0) {
		pvr_log_deinit();
		pvr_logfd = fd;

		return (0);
	}
	return (-1);
}

int
pvr_log_stdout(void)
{
	if (pvr_logfd < 0)
		return (-1);
	if (dup2(pvr_logfd, STDOUT_FILENO) < 0) {
		LOG("dup2() failed: %s", strerror(errno));
		return (-1);
	}

	return (0);
}

int
pvr_log_deinit(void)
{
	if (pvr_logfd >= 0) {
		close(pvr_logfd);
		pvr_logfd = -1;
	}
	return (0);
}

int
strlcpy(char *dst, const char *src, size_t dst_size)
{
	if (dst_size == 0)
		return (0);
	strncpy(dst, src, dst_size);
	dst[dst_size - 1] = '\0';

	return (0);
}

int
strlcat(char *dst, const char *src, size_t dst_size)
{
	if (dst_size == 0)
		return (0);
	strncat(dst, src, dst_size);
	dst[dst_size - 1] = '\0';

	return (0);
}

struct dirlist *
dirlist_get(char *dir, int flags)
{
	struct dirlist *list;
	struct direntry *entry, *newentry;
	struct dirent *ent;
	DIR *dirp;
	int old_errno;

	if ((dirp = opendir(dir)) == NULL)
		return (NULL);
	list = calloc(1, sizeof(*list));
	TAILQ_INIT(list);
	old_errno = errno;
	while ((ent = readdir(dirp)) != NULL) {
		/* Skip "." and ".." entries */
		if (ent->d_name[0] == '.' &&
			(ent->d_name[1] == '\0' ||
			 (ent->d_name[1] == '.' && ent->d_name[2] == '\0')))
			continue;

		/* Skip non-directories */
		if ((flags & DIRLIST_DIRONLY) && ent->d_type != DT_DIR)
			continue;

		newentry = calloc(1, sizeof(*newentry));
		strlcpy(newentry->name, ent->d_name, sizeof(newentry->name));
		newentry->d_type = ent->d_type;

		if (!(flags & DIRLIST_SORT))
			TAILQ_INSERT_TAIL(list, newentry, link);
		else {
			TAILQ_FOREACH(entry, list, link) {
				if (strcmp(entry->name, ent->d_name) < 0)
					continue;
				TAILQ_INSERT_BEFORE(entry, newentry, link);
				break;
			}
			if (entry == NULL)
				TAILQ_INSERT_TAIL(list, newentry, link);
		}
		old_errno = errno;
	}

	if (old_errno != errno) {
		LOG("Cannot read dir[%s]: %s", dir, strerror(errno));
		closedir(dirp);
		dirlist_free(list);
		return (NULL);
	}
	closedir(dirp);
	return (list);
}

int
dirlist_free(struct dirlist *list)
{
	struct direntry *entry;

	while ((entry = TAILQ_FIRST(list)) != NULL) {
		TAILQ_REMOVE(list, entry, link);
		free(entry);
	}
	free(list);
	return (0);
}

int dir_writable(char *dir)
{
	int fd;
	char path[MAX_PATH];

	snprintf(path, sizeof(path), "%s/%s", dir, TESTFILE);
	if ((fd = open(path, O_RDWR|O_CREAT|O_TRUNC, 0644)) < 0)
		return (0);
	close(fd);
	remove(path);
	return (1);
}
