#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "mod.h"
#include "mod_pvr.h"
#include "config.h"

static int fill_int(char *name, char *str, void *ptr, void *ctl);
static int fill_flag(char *name, char *str, void *ptr, void *ctl);
static int fill_string(char *name, char *str, void *ptr, void *ctl);
static int fill_enum(char *name, char *str, void *ptr, void *ctl);
static int fill_string_list(char *name, char *str, void *ptr, void *ctl);
static int fill_key(char *name, char *str, void *ptr, void *ctl);

/* fill configuration variables with default values */
cfgstruct config = {
	.debug = 1,
	.exedsp_debug = 0,
	.exedsp_facility = 0xffffffff,
	.exedsp_severity = 0x100000,

	.mod_key = KEY_GREEN,

	.default_source = 0x0,

	.aspect_ratio = 1,

	.audio_switch = 0,

	.log_keycodes = 0,

	.enable_esaving = 0,
	.esaving_key_check = MKEY|KEY_NUM0,
	.esaving_key_change = MKEY|KEY_NUM1,
	.esaving_key_mute = MKEY|KEY_NUM2,
	.esaving_key_panel_off = MKEY|KEY_NUM3,

	.path = "",
	.enable_pvr = 1,
	.pvr_target = TAILQ_HEAD_INITIALIZER(config.pvr_target),
	.record_key_lock = 1,
	.protect = 1,
	.unicode = CFG_UNICODE_UTF8,
	.winter_time = 1,
	.summer_time = 2,
	.show_icons = 1,
	.icon_x = 10,
	.icon_y = 10,
	.testwrite = 1,
	.del_mdb = 1,
	.del_empty_ts = 1,
	.io_buffer = 1732608,
};

struct pvr_enum_entry {
	char *name;
	int value;
};

static struct pvr_enum_entry ctl_unicode[] = {
	{"UTF",   CFG_UNICODE_UTF8},
	{"UTF-8", CFG_UNICODE_UTF8},
	{"UTF8",  CFG_UNICODE_UTF8},
	{"ASCII", CFG_UNICODE_ASCII},
	{"ASCII_TRANSCODE", CFG_UNICODE_ASCII_TRANSCODE},
	{NULL, 0}
};

static struct pvr_enum_entry ctl_default_source[] = {
	{"OFF",       0},
	{"TV",        KEY_SOURCE_TV},
	{"EXT1",      KEY_SOURCE_EXT1},
	{"EXT2",      KEY_SOURCE_EXT2},
	{"PC",        KEY_SOURCE_PC},
	{"AV",        KEY_SOURCE_AV},
	{"COMPONENT", KEY_SOURCE_COMPONENT},
	{"HDMI1",     KEY_SOURCE_HDMI1},
	{"HDMI2",     KEY_SOURCE_HDMI2},
	{"HDMI3",     KEY_SOURCE_HDMI3},
	{"DVI",       KEY_SOURCE_HDMI3},
	{"HDMI4",     KEY_SOURCE_HDMI4},
	{NULL, 0}
};

static struct pvr_enum_entry ctl_record_key_lock[] = {
	{"OFF",       RECORD_KEYLOCK_DISABLE},
	{"DISABLE",   RECORD_KEYLOCK_DISABLE},
	{"0",         RECORD_KEYLOCK_DISABLE},
	{"ENABLE",    RECORD_KEYLOCK_ENABLE},
	{"1",         RECORD_KEYLOCK_ENABLE},
	{"LOCKED",    RECORD_KEYLOCK_ENABLE_LOCKED},
	{"2",         RECORD_KEYLOCK_ENABLE_LOCKED},
	{"NOUNLOCK",  RECORD_KEYLOCK_NOUNLOCK},
	{"3",         RECORD_KEYLOCK_NOUNLOCK},
	{NULL, 0}
};

struct stringlist_ctl {
	char *delim;
	int (*check_func)(char *str);
};

static struct stringlist_ctl pvr_target_ctl = {":", pvr_check_target};

#define STR_ENTRY(str) fill_string, config.str, (void *)sizeof(config.str)
#define FLAG_ENTRY(flag) fill_flag, &config.flag, NULL
#define INT_ENTRY(integer) fill_int, &config.integer, NULL
#define ENUM_ENTRY(integer, ctl) fill_enum, &config.integer, ctl
#define STR_LIST_ENTRY(strlist, delim) fill_string_list, &config.strlist, (void *)delim
#define KEY_ENTRY(integer) fill_key, &config.integer, NULL

static struct pvr_config_var {
	char *name;
	int (*fill_cb)(char *name, char *str, void *ptr, void *ctl);
	void *ptr;
	void *ctl;
} vars[] = {
	{"LOG",             FLAG_ENTRY(debug)},
	{"EXEDSP_LOG",      FLAG_ENTRY(exedsp_debug)},
	{"EXEDSP_FACILITY", INT_ENTRY(exedsp_facility)},
	{"EXEDSP_SEVERITY", INT_ENTRY(exedsp_severity)},
	{"MOD_KEY",         KEY_ENTRY(mod_key)},
	{"DEFAULT_SOURCE",  ENUM_ENTRY(default_source, ctl_default_source)},
	{"ASPECT_RATIO",    KEY_ENTRY(aspect_ratio)},
	{"AUDIO_SWITCH",    KEY_ENTRY(audio_switch)},
	{"LOG_KEYCODES",    FLAG_ENTRY(log_keycodes)},
	{"ESAVING",         FLAG_ENTRY(enable_esaving)},
	{"ESAVING_CHECK",   KEY_ENTRY(esaving_key_check)},
	{"ESAVING_MUTE",    KEY_ENTRY(esaving_key_mute)},
	{"ESAVING_CHANGE",  KEY_ENTRY(esaving_key_change)},
	{"ESAVING_PANEL_OFF", KEY_ENTRY(esaving_key_panel_off)},
	{"PVR",             FLAG_ENTRY(enable_pvr)},
	{"TARGET",          STR_LIST_ENTRY(pvr_target, &pvr_target_ctl)},
	{"RECORD_KEY_LOCK", ENUM_ENTRY(record_key_lock, ctl_record_key_lock)},
	{"PROTECT",         FLAG_ENTRY(protect)},
	{"UNICODE",         ENUM_ENTRY(unicode, ctl_unicode)},
	{"WINTER_TIME",     INT_ENTRY(winter_time)},
	{"SUMMER_TIME",     INT_ENTRY(summer_time)},
	{"SHOW_ICONS",      FLAG_ENTRY(show_icons)},
	{"ICON_X",          INT_ENTRY(icon_x)},
	{"ICON_Y",          INT_ENTRY(icon_y)},
	{"TESTWRITE",       FLAG_ENTRY(testwrite)},
	{"DEL_MDB",         FLAG_ENTRY(del_mdb)},
	{"DEL_EMPTY_TS",    FLAG_ENTRY(del_empty_ts)},
	{"IO_BUFFER",       INT_ENTRY(io_buffer)},
	{NULL, NULL, NULL, NULL}
};

static char *
skip_spaces(char *ptr)
{
	while (isspace(*ptr))
		ptr++;
	return (ptr);
}

static int
fill_int(char *name, char *str, void *ptr, void *ctl)
{
	long long int var;

	if (sscanf(str, "%lli", &var) != 1)
		var = 0;

	if (var > 0xffffffff || var < 0)
		var = 0;
	*(unsigned *)ptr = var;
	LOG("%s = [%u / 0x%X]", name, *(unsigned *)ptr, *(unsigned *)ptr);
	
	return (0);
}

static int
fill_flag(char *name, char *str, void *ptr, void *ctl)
{
	int *flag = ptr;

	if (!strcasecmp(str, "yes") || !strcasecmp(str, "true") ||
		!strcasecmp(str, "enable") || !strcasecmp(str, "enabled") ||
		!strcasecmp(str, "on")) {

		*flag = 1;
	} else if (!strcasecmp(str, "no") || !strcasecmp(str, "false") ||
			   !strcasecmp(str, "disable") || !strcasecmp(str, "disabled") ||
			   !strcasecmp(str, "off")) {
		*flag = 0;
	} else if (*str == '\0')
		*flag = 0;
	else
		*flag = (atoi(str) > 0);

	LOG("%s = [%d]", name, *flag);

	return (0);
}

static int
fill_string(char *name, char *str, void *ptr, void *ctl)
{
	size_t size = (size_t)ctl;

	strlcpy((char *)ptr, str, size);
	LOG("%s = [%s]", name, (char *)ptr);

	return (0);
}

static int
fill_enum(char *name, char *str, void *ptr, void *ctl)
{
	struct pvr_enum_entry *entry = ctl;

	for (; entry->name != NULL; entry++) {
		if (strcasecmp(str, entry->name))
			continue;
		*(int *)ptr = entry->value;
		LOG("%s = [%d / 0x%X](%s)", name, entry->value, entry->value, str);
		return (0);
	}
	LOG("Unknown %s value[%s], leave the previous[%d / %X]",
		name, str, *(int *)ptr, *(int *)ptr);
	return (0);
}

static int
fill_string_list(char *name, char *str, void *ptr, void *ctl_)
{
	struct stringlist *list = ptr;
	struct stringlist_entry *entry;
	struct stringlist_ctl *ctl = ctl_;
	char *substr;

	while ((entry = TAILQ_FIRST(list)) != NULL) {
		TAILQ_REMOVE(list, entry, link);
		free(entry);
	}

	substr = strtok(str, ctl->delim);
	for (; substr != NULL; substr = strtok(NULL, ctl->delim)) {
		if (ctl->check_func != NULL && ctl->check_func(substr))
			continue;
		entry = calloc(1, sizeof(*entry));
		strlcpy(entry->value, substr, sizeof(entry->value));
		TAILQ_INSERT_TAIL(list, entry, link);
	}
	LOG("%s =", name);
	TAILQ_FOREACH(entry, list, link)
		LOG("  [%s]", entry->value);
	return (0);
}

static int
fill_key(char *name, char *str, void *ptr, void *ctl)
{
	char *str_orig = str;
	int key = 0;
	int rc = -1;
	struct {
		char ch;
		int flag;
		int present;
	} mods[] = {
		{'M', MKEY, 0},
		{'T', TKEY, 0},
		{0,   0,    0}
	}, *mod;

	if (!strcasecmp(str, "no") || !strcasecmp(str, "false") ||
		!strcasecmp(str, "disable") || !strcasecmp(str, "disabled") ||
		!strcasecmp(str, "off")) {

		rc = 0;
		goto DONE;
	}
    while (1) {
		if (str[0] != '\0' && str[1] == '+') {
			/* Try to parse modifier keys */
			char ch = toupper(str[0]);

			for (mod = mods; mod->ch != 0; mod++) {
				if (mod->present || ch != mod->ch)
					continue;
				/* ch == mod->ch */
				key |= mod->flag;
				mod->present = 1;
				str += 2;
				break;
			}
			if (mod->ch != 0)
				continue;
		}

		/* Try hexadecimal key value */
		if (str[0] == '0' && str[1] == 'x') {
			int tkey;
			int scanned;

			if (sscanf(str + 2, "%X%n", &tkey, &scanned) != 1)
				goto DONE;
			if (str[scanned + 2] != '\0')
				goto DONE;

			for (mod = mods; mod->ch != 0; mod++)
				if ((tkey & mod->flag))
					goto DONE;

			key |= tkey;
			rc = 0;
			goto DONE;
		} else {
			/* Try named key value */
			struct keyinfo *info;

			for (info = keyinfos; info->name != NULL; info++) {
				if (strcasecmp(info->name, str))
					continue;
				if (str[strlen(info->name)] != '\0')
					continue;
				key |= info->key;
				rc = 0;
				goto DONE;
			}
			goto DONE;
		}
		break;
	}

  DONE:
	if (rc != 0) {
		LOG("%s: cannot parse key[%s], leave the previous value[0x%X]",
			name, str_orig, *(int *)ptr);
	} else {
		*(int *)ptr = key;
		LOG("%s = [0x%X](%s)", name, key, str_orig);
	}
	return (0);
}

static int
set_default_config(void)
{
	struct stringlist_entry	*str_entry;
	
	/* TODO: move static cfgstruct structure initialization here */

	str_entry = calloc(1, sizeof(*str_entry));
	strlcpy(str_entry->value, "/dtv/usb/sda1", sizeof(str_entry->value));
	TAILQ_INSERT_TAIL(&config.pvr_target, str_entry, link);

	str_entry = calloc(1, sizeof(*str_entry));
	strlcpy(str_entry->value, "/mtd_wiselink/MOVIE001", sizeof(str_entry->value));
	TAILQ_INSERT_TAIL(&config.pvr_target, str_entry, link);

	return (0);
}

int
load_config(const char *path)
{
	struct pvr_config_var *var;
	FILE *cfg;
	char buf[MAX_PATH];
	char *name, *val, *ptr;
	int i;

	strlcpy(config.path, path, sizeof(config.path));

	/*
     * Initialize the logging. It can be desabled later, in
     * pvr_config().
     */
	snprintf(buf, sizeof(buf), "%s%s", config.path, "pvr.log");
	pvr_log_init(buf);
	LOG("Loading configuration file[%s]", buf);

	set_default_config();

	snprintf(buf, sizeof(buf), "%s%s", path, CONFIG_FNAME);
	if ((cfg = fopen(buf, "r")) == NULL) {
		LOG("Cannot open configuration file[%s]", buf);
		pvr_log_deinit();
		return (-1);
	}

	while (fgets(buf, sizeof(buf), cfg) != NULL) {
		for (i = 0; buf[i] != '\0'; i++) {
			if (buf[i] != '\r' && buf[i] != '\n')
				continue;
			buf[i] = '\0';
			break;
		}
		name = skip_spaces(buf);

		/* skip comment lines or lines with unknown format */
		if (strchr(";#/=", *name) != NULL) {
			continue;
		}
		if ((val = strchr(name, '=')) == NULL) {
			continue;
		}

		for (ptr = val - 1; ptr >= name && isspace(*ptr); ptr--);
		ptr[1] = '\0';

		for (var = vars; var->name != NULL; var++)
			if (!strcasecmp(var->name, name))
				break;
		if (var->name == NULL) {
 			LOG("Unknown variable[%s]",	name);
			continue;
		}

		val = skip_spaces(val + 1);
		for (ptr = val + strlen(val) - 1; ptr >= val && isspace(*ptr); ptr--);
		ptr[1] = '\0';

		var->fill_cb(name, val, var->ptr, var->ctl);
	}

	fclose(cfg);
	return (0);
}
