#define _GNU_SOURCE
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include "common.h"
#include "config.h"
#include "exedsp.h"
#include "epg.h"
#include "icons.h"
#include "mod.h"
#include "mod_pvr.h"

#define USB_DIR_TEMPLATE "$(USB)"
#define USB_DIR "/dtv/usb"
#define TESTFILE "pvr_testfile"
#define MAX_DEVICES 16

/* Number of Stop key presses before locking/unlocking */
#define RECORD_KEYLOCK_NUMBER 4

#define TS_FILE_PATTERN "%s/R%s%03d.ts"
#define MDB_FILE_PATTERN "%s/R%s%03d.mdb"

static char ts_filename[MAX_PATH] = "";
static char mdb_filename[MAX_PATH] = "";
static char recpartname[256];
static char target[MAX_PATH];
static int recording = 0;
static int record_locked = -1;

static void *pvr_rec = 0;

static void set_io_buffer_size(unsigned val)
{
   static unsigned *pvr_recinstance;

   if (val > 0)
   {
      spIPvr_GetRecInstance(0, &pvr_recinstance);
      if (pvr_recinstance) *(pvr_recinstance + 0x2C/4) = val;
   }
}

int
pvr_check_target(char *dir)
{
	size_t len = strlen(dir);

	if (len == 0) {
		LOG("Empty dir[%s]", dir);
		return (-1);
	}

	for (; len > 0 && dir[len - 1] == '/'; len--)
		dir[len - 1] = '\0';

	if (*dir == '/')
		return (0);

	if (strcasestr(dir, USB_DIR_TEMPLATE) == dir) {
		char *ptr = dir + (sizeof(USB_DIR_TEMPLATE) - 1);

		if (*ptr != '\0' && *ptr != '/') {
			LOG("Invalid template format for dir[%s]", dir);
			return (-1);
		}
		return (0);
	}
	LOG("Non-absolute path[%s]", dir);
	return (-1);
}

static int
pvr_find_usb_dir(char *postfix, char *buf, size_t siz)
{
	struct dirlist *list;
	struct direntry *entry;
	char path[MAX_PATH];
	int rc = -1;

	if ((list = dirlist_get(USB_DIR, DIRLIST_SORT|DIRLIST_DIRONLY)) == NULL)
		return (-1);

	TAILQ_FOREACH(entry, list, link) {
		if (entry->name[0] != 's' || entry->name[1] != 'd' ||
			entry->name[2] < 'a' || entry->name[2] >= 'a' + MAX_DEVICES)
			continue;
		if (*postfix != '\0')
			snprintf(path, sizeof(path), "%s/%s/%s", USB_DIR, entry->name, postfix);
		else
			snprintf(path, sizeof(path), "%s/%s", USB_DIR, entry->name);
		if (dir_writable(path)) {
			strlcpy(buf, path, siz);
			rc = 0;
			goto DONE;
		}
		LOG("Dir[%s] is not writable", path);
	}

  DONE:
	dirlist_free(list);
	return (rc);
}

static int
pvr_select_target_dir(char *buf, size_t siz)
{
	struct stringlist_entry *entry;
	char *ptr;
	

	TAILQ_FOREACH(entry, &config.pvr_target, link) {
		if (strcasestr(entry->value, USB_DIR_TEMPLATE) == entry->value) {
			ptr = entry->value + (sizeof(USB_DIR_TEMPLATE) - 1);
			if (*ptr == '/')
				ptr++;
			if (pvr_find_usb_dir(ptr, buf, siz))
				continue;
			break;
		}
		if (!dir_writable(entry->value)) {
			LOG("Dir[%s] is not writable", entry->value);
			continue;
		}
		strlcpy(buf, entry->value, siz);
		break;
	}
	if (entry != NULL) {
		LOG("Selected target dir[%s]", buf);
		return (0);
	}
	LOG("Cannot find target directory");
	return (-1);
}

static int create_recname()
{
	FILE *testwrite;
	int nr;

	if (pvr_select_target_dir(target, sizeof(target)) != 0)
		return (-1);

	EPG_get_data(recpartname, sizeof(recpartname));
	LOG("Got epg data[%s]", recpartname);

	for (nr = 1; nr < 999; nr++) {
		snprintf(ts_filename, sizeof(ts_filename),
				 "%s/R%s%03d.ts", target, recpartname, nr);
		if (access(ts_filename, F_OK) != 0) {
			if (config.testwrite) {
				testwrite = fopen(ts_filename, "a+");
				if (testwrite) {
					fclose(testwrite);
					remove(ts_filename);
				} else
					continue;
			}
			sprintf(mdb_filename, "%s/R%s%03d.mdb",
					target, recpartname, nr);
			return nr;
		}
	}
	return (-1);
}

/*
 * Bugged spI_Stream_Record() replacement.
 *
 * To see how to debug spI_Stream_Record(), see revision 14:18599fa806c5
 * of http://bitbucket.org/sorcerer/samygopvr
 */
static int
pvr_stream_record(char *filename)
{
	struct pvr_rec_set_info record_info;
	int fname_len;
	int fd = -1;
	int rc = -1;

	fname_len = strlen(filename);
	if (fname_len >= sizeof(record_info.fname)) {
		LOG("Too long filename[%s]: %d >= %d",
			filename, fname_len,
			(int)sizeof(record_info.fname));
		goto DONE;
	}

	bzero(&record_info, sizeof(record_info));
	strlcpy(record_info.fname, filename, sizeof(record_info.fname));

	record_info.video_pid = 1;
	record_info.audio_pid = 2;
	record_info.pcr_pid = 3;

	/*
     * Fill record_info structure with magic values got from
     * pvr_rec_set_test(). I think most of the big numbers are just
     * uninitialized variables of spI_Stream_Record() but they
     * shouldn't be initialized with zero or spIPvr_RecSet() will
     * fail.
     */
	record_info.xC8 = 1;
	record_info.xC4 = 0;
	record_info.xC0 = 0;
	record_info.xBC = 0;
	record_info.x5C = 0;
/* 	record_info.x30 = 1; */
	record_info.x30 = 0x71E113EC;
	record_info.xB8 = 1;

	record_info.unk2[0] = 0;
	record_info.unk2[1] = 0x1000;
	record_info.unk3[0] = 0x40884000;
	record_info.unk3[1] = 0x2;
	record_info.unk3[2] = 0x407CF8B8;
	record_info.unk5[0] = 0xFFFFFFFF;
	record_info.unk5[1] = 0;
	record_info.x9C = 0;
/* 	record_info.x64 = 0; */
 	record_info.x64 = 1;
	record_info.unk6 = 0x2D64550;
	record_info.x7C = 0;
/* 	record_info.x70 = 0; */
	record_info.x70 = 0x2D64550;
	record_info.unk7[0] = 1;
	record_info.unk7[2] = 0x1008;
	record_info.unk7[3] = 0x40885054;
	record_info.unk7[4] = 0x2D645F0;
	record_info.unk7[5] = 0x40884000;
	record_info.unk7[6] = 0x40883D10;
	record_info.unk7[9] = 0x7170E5DC;
	record_info.unk8[0] = 0x1A;
	record_info.unk8[1] = 0x71E11B01;
	record_info.unk8[2] = 0x71E11C2C;
	record_info.unk8[3] = 0x49;

	if ((fd = open(record_info.fname, O_RDWR | O_CREAT, 0666)) < 0) {
		LOG("Cannot open file[%s]: %s", record_info.fname, strerror(errno));
		goto DONE;
	}
	if ((rc = spIPvr_RecOpen(0, &pvr_rec)) != 0) {
		LOG("Cannot open PVR rec device: %x", rc);
		goto DONE;
	}
	LOG("Opened PVR rec device[%p]", pvr_rec);
	if ((rc = spIPvr_RecSet(pvr_rec, &record_info, 1)) != 0) {
		LOG("Cannot set PVR rec device[%p]: %X", pvr_rec, rc);
		spIPvr_RecClose(pvr_rec);
		goto DONE;
	}
	LOG("Set PVR rec device[%p]", pvr_rec);
	if ((rc = spIPvr_RecStart(pvr_rec)) != 0) {
		LOG("Cannot start PVR rec device[%p]: %X", pvr_rec, rc);
		spIPvr_RecClose(pvr_rec);
		goto DONE;
	}
  DONE:
	if (rc == 0)
		LOG("Started recording to[%s] using PVR rec device[%p]",
			record_info.fname, pvr_rec);
	if (fd != -1)
		close(fd);
	return (rc == 0) ? 0 : -1;
	
}

/* spI_Stream_Record_Stop() replacement */
static int
pvr_stream_record_stop(void)
{
	int rc;

	if (pvr_rec == NULL) {
		LOG("Not in record state!");
		return (-1);
	}
	if ((rc = spIPvr_Rec_UnregisterCallback(pvr_rec, 1)) != 0) {
		LOG("Failed to unregister record stream callback "
			"on device[%p]: %x", pvr_rec, rc);
		return (-1);
	}
	if ((rc = spIPvr_RecStop(pvr_rec)) != 0) {
		LOG("Failed to stop PVR record device[%p]: %x", pvr_rec, rc);
		return (-1);
	}
	if ((rc = spIPvr_RecClose(pvr_rec)) != 0) {
		LOG("Failed to close PVR record device[%p]: %x", pvr_rec, rc);
		return (-1);
	}
	LOG("Stopped recording with PVR rec device[%p]", pvr_rec);
	pvr_rec = NULL;

	return (0);
}

/* Check for RECORD_KEYLOCK_NUMBER stop key presses */
static int
pvr_check_record_lock(const int *keys, int *key_out)
{
	static int stop_count = 0;

	if (record_locked == -1) {
		/* Initialize record_key_lock function variables */
		if (config.record_key_lock == RECORD_KEYLOCK_ENABLE ||
			config.record_key_lock == RECORD_KEYLOCK_NOUNLOCK)
			record_locked = 0;
		else
			record_locked = 1;
	}

	if (keys[0] != KEY_STOP) {
		stop_count = 0;
		return (MOD_KEYPRESSED_UNKNOWN);
	}

	if (config.record_key_lock != RECORD_KEYLOCK_DISABLE &&
		(config.record_key_lock != RECORD_KEYLOCK_NOUNLOCK ||
		 !record_locked)) {

		stop_count++;
		if (stop_count == RECORD_KEYLOCK_NUMBER) {
			stop_count = 0;
			record_locked = !record_locked;
			show_WARNING(record_locked ? ICON_LOCKED : ICON_UNLOCKED);
			*key_out = KEY_NOTHING;
			return (MOD_KEYPRESSED_PROCESSED);
		}
	}
	return (MOD_KEYPRESSED_UNKNOWN);
}

int
pvr_keypressed(const int *keys, int keys_num, int *key_out)
{
	struct stat st;
	int key = keys[0];
	int deleted;
	int rc;
	int ret = MOD_KEYPRESSED_UNKNOWN;

	if ((ret = pvr_check_record_lock(keys,
									 key_out)) == MOD_KEYPRESSED_PROCESSED)
		return (ret);

	switch (key) {
	case KEY_STOP:
	case KEY_POWER:
		if (!recording) {
			LOG("We're not recording");
			break;
		}
		ret = MOD_KEYPRESSED_PROCESSED;
		recording = 0;


#if 0
		spI_Stream_Record_Stop();
		rc = 0;
#endif
		rc = pvr_stream_record_stop();
		LOG("recording stopped, rc[%d]", rc);

		if (mdb_filename[0]) {
			if (config.del_mdb) {
				remove(mdb_filename);
				LOG("Removed mdb file[%s]", mdb_filename);
			}
			mdb_filename[0]=0;
		}

		deleted = 0;
		if (ts_filename[0] && stat(ts_filename, &st) == 0) {
			if (config.del_empty_ts && st.st_size == 0) {
				remove(ts_filename);
				deleted = 1;
				LOG("Removed empty ts file[%s]", ts_filename);
			} else
				LOG("Written %"PRIu64" bytes to file[%s]",
					(uint64_t)st.st_size, ts_filename);
		}

		if (key == KEY_STOP)
			*key_out = hide_REC(deleted);
		break;

	case KEY_RECORD:
		ret = MOD_KEYPRESSED_PROCESSED;
		if (recording) {
			*key_out = show_FORBIDDEN();
			LOG("We're already recording, dir[%s], ret[0x%X]",
				target, key);
		} else if (config.record_key_lock != RECORD_KEYLOCK_DISABLE &&
				   record_locked) {
			*key_out = show_WARNING(ICON_LOCKED);
			LOG("Recording is locked, ret0x%X]", key);
		} else {
			int cn = create_recname();

			if (cn > 0) {

#if 0
				spI_Stream_Record(ts_filename);
#endif
				if ((rc = pvr_stream_record(ts_filename)) != 0) {
					LOG("Cannot record to[%s]", ts_filename);
					*key_out = show_WARNING(ICON_NOREC);
				} else {
					recording = 1;

					LOG("Started recording to[%s], rc[%d] ret[0x%X]",
						ts_filename, rc, key);
					*key_out = show_REC();
				}
			}
			else {
				*key_out = show_WARNING(ICON_NOREC);
				LOG("Cannot recording, cn[%d] ret[0x%X]", cn, key);
			}
		}
		break;

	case KEY_PAUSE:
		ret = MOD_KEYPRESSED_PROCESSED;
		if (recording)
			*key_out = icon_displayed()?hide_REC(0):show_REC();
		break;

	case KEY_SOURCE:
	case KEY_P_UP:
	case KEY_P_DOWN:
	case KEY_PRE_CH:
	case KEY_CH_LIST:
	case KEY_SOURCE_TV:
	case KEY_NUM0:
	case KEY_NUM1: case KEY_NUM2: case KEY_NUM3:
	case KEY_NUM4: case KEY_NUM5: case KEY_NUM6:
	case KEY_NUM7: case KEY_NUM8: case KEY_NUM9:
	case KEY_CONTENT:
	case KEY_MEDIA_P:
	case KEY_TTX_MIX:
	case KEY_FAV_CH:
		if (recording && config.protect) {
			LOG("Forbidden key pressed[0x%X]", key);
			*key_out = show_FORBIDDEN();
			ret = MOD_KEYPRESSED_PROCESSED;
		}
		break;

	default:
		break;
	}

	return (ret);
}

int
pvr_init(void)
{
	if (!config.enable_pvr)
		return (MOD_INIT_IGNORE);
	set_io_buffer_size(config.io_buffer);
	return (MOD_INIT_DONE);
}
